Require Import Classical.
Require Import Program.
Require Import List.
Require Import Omega.
Import ListNotations.

Module Type SigMod.

Parameter State : Type.
Parameter trans : State -> State -> Prop.

End SigMod.



Module   LogicMod(ST : SigMod).
Import ST.

Definition SymState := State -> Prop.
Definition top : SymState := fun _ => True.
Definition bot : SymState := fun _ => False.
Definition cnj(p q :SymState) := fun s => p s /\ q s.
Definition dsj(p q :SymState) := fun s => p s \/ q s.
Definition neg (p:SymState) := fun s => ~ p s.
Definition imp(p q :SymState) := forall s, p s -> q s.
(* symbolic transition function *)
Definition symTrans(q:SymState) := fun s => exists s', q s' /\ trans s' s.

(* symbolic transition relatiion is monotonous *)
Lemma symTrans_mono : forall p q,  imp p q -> imp (symTrans p) (symTrans q).
Proof.
unfold imp, symTrans ; intros.
destruct H0 as [s' [Hp Htr]].
exists s'; split ; auto.
Qed.

(* final states *)
Definition final : SymState := fun s => forall s', ~trans s s'.

(* sequences and paths *)
CoInductive seq  : Type :=
| one : forall s:State , seq
| add : State -> seq  -> seq.                                          
Definition head(sq : seq) : State :=
  match sq with
    | one s => s
    | add s _ => s
   end.


 CoInductive isPath: seq -> Prop :=
|path_one : forall s,  final s -> isPath (one s)
|path_add : forall s sq,  trans s (head sq) -> isPath sq -> isPath (add s sq).

(* when does a sequence "reach" a symbolic state *)
CoInductive reach : seq -> SymState -> Prop :=
|reach_now_one : forall s (r:SymState),  r s-> reach (one s) r
|reach_now_add : forall s tau (r:SymState),   r s-> reach (add s tau) r
|reach_later : forall s tau (r:SymState),  reach tau r -> reach (add s tau) r.



(* reachability-logic forrmulas*)
Inductive Rlf:Type  :=
| rlf: forall p q: SymState, Rlf.

Notation "l  =><> r"  := (rlf  l r) (at level 100).                              

Definition lhs (f : Rlf) := match f with (l  =><> _)  => l end.
Definition rhs (f : Rlf) := match f with (_  =><> r)  => r end.



(*validity*)
Definition valid(f:Rlf) : Prop :=
  forall tau,   isPath tau ->  (lhs f) (head tau) -> reach tau (rhs f).


(* a one-rule proof system for infinite symbolic execution *)
CoInductive proof:  Rlf  -> Prop :=
|Stp : forall (l l'  r: SymState),
    imp l (dsj l' r) -> (* i.e., l' over-approximates l \ r*)
    imp (cnj l' final) bot ->
    proof  (symTrans l' =><> r) ->
    proof (l =><> r).

(* soundness: all proved rlfs are valid*)
Theorem soundness : forall f, proof f -> valid f.
Proof.
  cofix soundness. (*introducing the coinductive hypothesis *)
  intros f Hproof.
  inversion Hproof ; subst.
  intros tau HisPath Hinit ; cbn in *.
inversion HisPath ; subst.
* constructor.
  destruct (H  _ Hinit); try assumption.
    exfalso.
    inversion HisPath ; subst.
    assert (Hbot : bot s) ; intuition.
    apply H0 ; split ; assumption.
*  destruct (classic (r s)).
    + constructor ; unfold imp ; intros ; subst;  assumption.
    + constructor 3 ; try assumption.
      assert(Hval : valid(symTrans l' =><>  r)).
      { apply soundness ; auto. } (* coinductive hypothesis applied here*)
       apply Hval ; auto.
       exists s ; split ; auto.
       specialize (H _ Hinit).
       destruct H ; auto ; intuition.
Defined.  



Section ProvingSomeRLFormulas.
(* let's assume three symbolic states l, q, r *)  
  Variable l  r  q : SymState.
(* if the folllowing hypotheses hold on l, q and r, then (l =><> r) has a proof and thus is valid *)
  Hypothesis Hinit : imp  l (dsj q r).
  Hypothesis Hnonblock : imp (cnj q final) bot.
  Hypothesis HquasiInv : imp (symTrans q) (dsj q r).

Lemma strategy : proof (l =><> r).
Proof.
  apply Stp with q; auto.  
  cofix strategy.
  apply Stp with q; auto.
Defined.


End ProvingSomeRLFormulas.


(* we now prove that a "q" llike above always exists for valid rechability formulas *)
Definition coReach(r:SymState) : SymState :=
  fun s => ~ r s /\ forall tau, s = head tau-> isPath tau -> reach tau r.

Lemma init : forall f, valid f -> imp (lhs f) (dsj (coReach (rhs f)) (rhs f)).
Proof.
  unfold valid, coReach, imp, dsj; intros.
  destruct f as (l,r) ; cbn in *.
  destruct (classic (r s)) ; auto.
  left ; split; auto ; intros ; subst.
   apply H; auto.    
Defined.

Lemma nonBlock : forall r,  imp (cnj (coReach r) final) bot.
Proof.
  unfold imp, cnj, coReach, bot ; intros.
  do 2 destruct H.
  assert (isPath (one s)).
  * constructor ; auto.
  *  specialize (H1 (one s) (refl_equal _) H2).
     inversion H1; subst ; intuition.
Defined.

Lemma quasiInv : forall r, imp (symTrans (coReach r)) (dsj  (coReach r) r).
Proof.
  unfold imp, symTrans, coReach, dsj; intros.
  destruct H as [s' [[Hnr Hreach] Htrans]].
  destruct (classic (r s)) ; auto.
  left  ; split ; auto ; intros ; subst.
  assert (reach (add s' tau) r).
  * apply Hreach ; try reflexivity.
    constructor ; auto.
  *  inversion H0 ; subst ; intuition.
Defined.

Theorem completeness : forall f,  valid f -> proof f.
  intros.
  destruct f as (l,r).
  apply strategy with (coReach r).
  * apply (init (l =><> r))  ; auto.
  *  apply nonBlock.
  *  apply quasiInv.
 Defined.    

End LogicMod.