Require Import Classical.
Require Import Program.
Require Import List.
Require Import Omega.
Import ListNotations.
(* Coq 8.9 with paco package required *)
Require Import Paco.paco.

Module Type SigMod.

Parameter State : Type.
Parameter trans : State -> State -> Prop.

End SigMod.


Module   LogicMod(ST : SigMod).
Import ST.

(*symbolic states *) 
Definition SymState := State -> Prop.
Definition top : SymState := fun _ => True.
Definition bot : SymState := fun _ => False.
Definition cnj(p q :SymState) := fun s => p s /\ q s.
Definition dsj(p q :SymState) := fun s => p s \/ q s.
Definition neg (p:SymState) := fun s => ~ p s.
Definition imp(p q :SymState) := forall s, p s -> q s.

(* symbolic transition function *)
Definition symTrans(q:SymState) := fun s => exists s', q s' /\ trans s' s.

(* final states *)
Definition final : SymState := fun s => forall s', ~trans s s'.

(* sequences and paths *)
CoInductive seq  : Type :=
| one : forall s:State , seq
| add : State -> seq  -> seq.                                          
Definition head(sq : seq) : State :=
  match sq with
    | one s => s
    | add s _ => s
   end.

Inductive isFinite : seq -> Prop:=
|is_finite_one : forall s, isFinite (one s)
|is_finite_add : forall s sq, isFinite sq -> isFinite (add s sq).                                     


Fixpoint nth(n:nat)(sq:seq) : option State :=
  match n,sq with
  |0,  one s => Some s
  |0,  add s _  => Some s                    
  |S n', add _ tau' => nth n' tau'
  |_,_ => None
 end.           

Fixpoint suffix(n:nat)(sq:seq) : option seq :=
  match n,sq with
  |S n', add _ tau' => suffix n' tau'   
  |0, _ => Some sq
  |_,_ => None
 end.           

Lemma nth_suffix_nth  :
  forall n m s tau tau', suffix n tau = Some tau' -> nth m tau' = Some s ->
  nth (n + m) tau = Some s.                                             
Proof.
 induction n ; intros.
* cbn in *. injection H ; intros ; subst ; assumption.
* destruct tau ; cbn in * ; try discriminate.
   eapply IHn ; eauto.  
Qed.

Lemma nth_Some_suffix : forall n sq s',  nth n sq = Some s' ->  exists sq', suffix n sq = Some sq'.
Proof.
induction n  ; intros.
* destruct sq ; cbn in *.
  + injection H ; intros ; subst.  
    now exists (one s').  
  + injection H ; intros ; subst.  
      now exists (add s' sq).  
 * destruct sq ; cbn in *  ; try discriminate.
    eapply IHn ; eauto.
Qed.


Lemma head_nth_suffix : forall n sq  sq' s,
    suffix n sq = Some sq' ->
    nth n sq = Some s ->
    s = head sq'.
Proof.
induction n ; intros.
 * destruct sq ; cbn in *.
    +injection H0 ; injection H ; intros  ; subst ; reflexivity.
    +injection H0 ; injection H ; intros  ; subst ; reflexivity.
*  destruct sq ; cbn in * ; try discriminate.
   eapply IHn; eauto.
Qed.

Lemma isFinite_suffix : forall n sq sq' , isFinite sq -> suffix n sq = Some sq' -> isFinite sq'.
Proof.
 induction n ; intros.
 *   destruct sq ; cbn in *.
    +  injection H0 ; intros ; subst; constructor.
    + injection H0 ; intros ; subst; constructor.
       inversion H ; auto.      
*  destruct sq ; cbn in * ; try discriminate.
   inversion H; subst.
   eapply IHn ; eauto.   
Qed.

 CoInductive isPath: seq -> Prop :=
|path_one : forall s,  final s -> isPath (one s)
|path_add : forall s sq,  trans s (head sq) -> isPath sq -> isPath (add s sq).

Lemma isPath_suffix : forall n sq  sq' ,  isPath sq -> suffix n sq = Some sq' -> isPath sq'.
Proof.
 induction n ; intros.
* destruct sq; cbn in *.
   + injection H0 ; intros ; subst.
      inversion H0 ; subst ; auto.
   + injection H0 ; intros ; subst ; auto.
 *   destruct sq; cbn in *.
     + discriminate.
     + inversion H ; subst.
       apply IHn with sq ; auto.
Qed.

Lemma suffix_n_add : forall n tau tau' s, suffix n tau = Some tau' ->
  exists s',  suffix n (add s tau) = Some (add s' tau').
Proof.
induction n ; intros; cbn in*.
* injection H ; intros; subst ;auto.
  now exists s.
*destruct tau ; try discriminate.
 apply IHn ; auto.
Qed.

Inductive reach_gen rho:   seq -> SymState-> Prop :=
|_reach_now_one : forall  s (r:SymState),  r s-> reach_gen rho (one s) r
|_reach_now_add : forall  s tau (r:SymState),   r s-> reach_gen rho (add s tau) r
|_reach_later : forall s tau n tau'(r:SymState) (_: rho tau' r:Prop),
    suffix n tau = Some tau' -> reach_gen rho (add s tau) r.

Lemma reach_gen_mon: monotone2 reach_gen.
Proof.
  unfold monotone2 ; intros.
  inversion IN; subst ; cbn.
  * constructor ; auto.
  * constructor ; auto.
  * econstructor 3 ; eauto.
Qed.

Definition reach := paco2 reach_gen bot2.


Lemma suffix_reach : forall n tau tau' r,
  reach tau' r -> suffix n tau = Some tau' -> reach tau r.
Proof.
induction n ; intros.
* injection H0 ; intros ; subst ; auto.
* destruct tau ; try discriminate.
  pfold ; constructor 3 with 0 tau ; auto.
  left ; eapply IHn ; eauto. 
Qed.



Inductive reach_gen' rho:   seq -> SymState-> Prop :=
|_reach_now_one' : forall  s (r:SymState),  r s-> reach_gen' rho (one s) r
|_reach_now_add' : forall  s tau (r:SymState),   r s-> reach_gen' rho (add s tau) r
|_reach_later' : forall s tau r (_: rho tau r:Prop), reach_gen' rho (add s tau) r.



Lemma reach_gen'_mon: monotone2 reach_gen'.
Proof.
  unfold monotone2 ; intros.
  inversion IN; subst ; cbn.
  * constructor ; auto.
  * constructor ; auto.
  * econstructor 3 ; eauto.
Qed.

Definition reach' := paco2 reach_gen' bot2.


Lemma reach_reach' : forall tau r, reach tau r -> reach' tau r.
Proof.
  pcofix reach_reach' with rho.
  intros tau r Hreach.
  pinversion Hreach ; subst; try apply reach_gen_mon.
  *pfold ; constructor ; auto.
  *pfold ; constructor 2 ; auto.
  * pfold ; constructor 3.
    right.
    apply reach_reach'.
    apply suffix_reach with n tau' ; auto.
Qed.


Lemma reach'_reach : forall tau r, reach' tau r -> reach tau r.
Proof.
  pcofix reach'_reach with rho.
  intros tau r Hreach.
  pinversion Hreach ; subst ; try apply reach_gen'_mon.
  *pfold ; constructor ; auto.
  * pfold ; constructor 2 ; auto.
  * pfold; constructor 3 with 0 tau0.
    + right.
      apply reach'_reach ; auto.
    + reflexivity.  
Qed.


Lemma isFinite_reach_ex : forall sq r,
    isFinite sq ->reach sq r -> exists n s,  nth n sq = Some s /\ r s.
Proof.
  intros sq r Hfin Hreach.
  induction Hfin.
  apply reach_reach' in Hreach.
  * pinversion Hreach ; intros ; subst.
    exists 0,  s ; split ; auto.
     apply reach_gen'_mon.
  * pinversion Hreach ; intros ; subst.
     + inversion Hreach; subst.
        - exists 0,  s ; cbn ; split ; auto.
        - destruct H1.
          ** assert (Hreach1 : reach sq r).
              {
                apply suffix_reach with n tau' ; auto.
              }        
            specialize  (IHHfin Hreach1).
               destruct IHHfin as [n' [s' [Hnth Hr ]]].     
               now exists (S n'),  s'.
          ** inversion H.    
      +  ** assert (Hreach1 : reach sq r).
              {
                apply suffix_reach with n tau' ; auto.
              }  
       specialize  (IHHfin Hreach1).
       destruct IHHfin as [n' [s' [Hnth Hr ]]].     
       now exists (S n'),  s'.
     +  apply reach_gen_mon.
Qed.

Lemma nth_reach :   forall n sq (r:SymState) s,
   nth n sq = Some s -> r s -> reach sq r.
Proof.
  induction n  ; intros;  apply reach'_reach.  
* destruct sq  ; cbn in *.
+ injection H ; intros ; subst.
  pfold.
  constructor; auto.
+ injection H ; intros ; subst.
  pfold ; constructor ; auto.
 *  destruct sq  ; cbn in * ; try discriminate.
    pfold ; constructor 3; auto.
    left.
    apply reach_reach' ; eapply IHn ; eauto.
Qed.


(* reachability formulas*)
Inductive Rlf:Type  :=
| rlf: forall p q: SymState, Rlf.

Notation "l  =><> r"  := (rlf  l r) (at level 100).                              

Definition lhs (f : Rlf) := match f with (l  =><> _)  => l end.
Definition rhs (f : Rlf) := match f with (_  =><> r)  => r end.



(*validity*)

Definition valid(f:Rlf) : Prop :=
  forall tau,   isPath tau ->  isFinite tau -> (lhs f) (head tau) -> reach tau (rhs f).

Definition valid_gen rho f := forall tau, isPath tau ->isFinite tau -> (lhs f) (head tau) ->
  paco2 reach_gen rho tau (rhs f).

(*some properties of validity *)


Lemma valid_gen_trv : forall r rho,  valid_gen rho (r =><> r).
 unfold valid_gen ; cbn ; intros.
  pfold.
destruct tau; constructor ; auto.
Qed.


Lemma valid_gen_str : forall l l' r rho,
    imp l' l -> valid_gen rho (l =><> r) -> valid_gen rho ( l'=><> r).
Proof.
  unfold valid_gen ; cbn ; intros.
  apply H0, H ; auto.
Qed.

Lemma valid_gen_dsj : forall l1 l2 r rho,
    valid_gen rho (l1 =><> r) -> valid_gen rho (l2 =><>r) -> valid_gen rho ((dsj l1 l2) =><>  r).
Proof.
  unfold valid_gen  ; cbn ; intros.
  destruct H3.
  * apply H; assumption.
  * apply H0; assumption.  
Qed.


Lemma valid_finite_ex :
  forall l r,  valid ( l =><> r) ->
               forall tau,  l (head tau) -> isPath tau-> isFinite tau -> exists n s,  nth n tau = Some s /\ r s.
intros. 
unfold valid in *; cbn in *.
specialize (H _ H1 H2 H0).
specialize (isFinite_reach_ex  _ _ H2 H)  ; intro Hex.
 destruct Hex as [n [s [Hnth Hr]]]. 
now exists n, s.
Qed.


Lemma ex_finite_valid :
  forall (l r:SymState),
    (forall  tau,  l (head tau) -> isPath tau-> isFinite tau -> exists n s,  nth n tau = Some s /\ r s)
    ->
    valid ( l =><> r).
Proof.
  unfold valid ; intros ; cbn in *.
  specialize (H _ H2 H0 H1).
  destruct H as [n [s [Hnth Hr]]].    
    eapply nth_reach ; eauto.
Qed.


Lemma valid_iff_finite_ex :  forall (l r:SymState),
       valid ( l =><> r) <->   (forall  tau,  l (head tau) -> isPath tau-> isFinite tau -> exists n s,  nth n tau = Some s /\ r s).
Proof.
split.
 * apply valid_finite_ex.  
 * apply ex_finite_valid.
Qed.


Lemma valid_tra : forall l r m,  valid(l =><> m) -> valid (m =><> r) -> valid(l =><> r).
Proof.
  intros.
  rewrite valid_iff_finite_ex in * ; intros.
 specialize (H _ H1 H2 H3).
 destruct H as [n [s' [Hnth Hm]]].
 specialize (nth_Some_suffix _ _ _ Hnth) ; intro Hexsuf.
 destruct Hexsuf as [tau' Hsuf]. 
 specialize (head_nth_suffix _ _ _ _ Hsuf Hnth) ; intro Hhds'  ; subst.
specialize (isFinite_suffix _ _ _ H3 Hsuf) ; intro Hfinsuf.
specialize (isPath_suffix  _ _ _ H2 Hsuf) ; intro Hpathsuf.
specialize (H0 _ Hm Hpathsuf Hfinsuf).
 destruct H0 as [n' [s [Hnth' Hr]]].
exists (n + n'), s ; split ; auto.
eapply nth_suffix_nth ; eauto. 
Qed.


Lemma valid_halfgen_tra : forall rho l m r,
    valid(l =><> m) ->
    valid_gen rho (m =><> r) ->
    valid_gen rho (l =><> r).
Proof.
intros   rho l m r Hval Hvalgen tau HisPath HisFinite Hinit; cbn in *.
rewrite valid_iff_finite_ex in Hval.
specialize (Hval _ Hinit HisPath HisFinite).
 destruct Hval as [n [s' [Hnth Hm]]].
 specialize (nth_Some_suffix _ _ _ Hnth) ; intro Hexsuf.
 destruct Hexsuf as [tau' Hsuf]. 
 specialize (head_nth_suffix _ _ _ _ Hsuf Hnth) ; intro Hhds'  ; subst.
specialize (isFinite_suffix _ _ _ HisFinite Hsuf) ; intro Hfinsuf.
specialize (isPath_suffix  _ _ _ HisPath Hsuf) ; intro Hpathsuf.
unfold valid_gen in Hvalgen.
specialize (Hvalgen _ Hpathsuf Hfinsuf Hm) ; cbn in *.
pfold.
inversion HisFinite ; subst.
* destruct n ; try discriminate.
  injection Hsuf ; intros ; subst ; cbn in *.
  punfold Hvalgen; try apply reach_gen_mon.
*  destruct n ; cbn in *.  
   + injection Hsuf ; intros ; subst ; cbn in *.
     punfold Hvalgen; try apply reach_gen_mon.
   + constructor 3 with n tau' ; auto.
Qed.


Inductive proof_gen rho  : Rlf -> Prop :=
|Trv : forall r,  proof_gen rho ( r =><> r)
|Str : forall l l' r,
    imp l l' ->
    proof_gen rho (l' =><> r)  ->
    proof_gen rho  (l =><> r)               
|Spl : forall l1 l2 r,
    proof_gen rho (l1 =><> r) ->
    proof_gen rho (l2 =><> r) ->
    proof_gen rho ((dsj l1 l2) =><> r)
|Tra :
    forall l m r,
    valid (l =><> m) ->
    proof_gen rho (m =><> r) ->
    proof_gen rho (l =><> r)
|Stp : forall l r (_ : rho (symTrans l =><> r):Prop),
    imp (cnj l final) bot ->
    proof_gen rho (l =><> r).       


Lemma proof_gen_mon : monotone1 proof_gen.
Proof.
unfold monotone1 ; intros.
induction IN.
* constructor.
* econstructor 2 ; eauto.
* constructor 3 ; auto.
* econstructor 4 ; eauto.
* econstructor 5 ; eauto. 
Qed.

Definition proof := paco1 proof_gen bot1.


Theorem soundness : forall f, proof f -> valid f.
Proof.  
  pcofix soundness.
  assert (Hvalidgen: forall f,  proof f -> valid_gen r f); auto.
  intros (l,r0) Hproof.
  punfold Hproof ;  try apply proof_gen_mon.
   induction Hproof; subst.
   * apply valid_gen_trv.
   * apply valid_gen_str with l' ; auto.
   * apply valid_gen_dsj ; auto.
   * apply valid_halfgen_tra with m ; auto.
   * inversion H ; clear H.
     +
      intros tau HisPath Hfin Hinit; cbn in *.
      inversion HisPath ; subst ; cbn in *.
      - exfalso ; apply (H0 s) ; split ; auto.
      - pfold.
     destruct (classic (r1  s)).
      ** constructor ; auto.
      ** econstructor 3 with 0 sq; auto.
         right.
         apply (soundness _  H1)  ; auto.
         ++ inversion Hfin; auto.
         ++ now exists s.
     + inversion H1.                   
Qed.

Section CompleteStrategy.
(* let's assume three symbolic states l, q, r *)  
  Variable l  r  q : SymState.
(* if the folllowing hypotheses hold on l, q and r, then (l =><> r) has a proof and thus is valid *)
  Hypothesis Hinit : imp  l (dsj q r).
  Hypothesis Hnonblock : imp (cnj q final) bot.
  Hypothesis HquasiInv : imp (symTrans q) (dsj q r).

  

  
Lemma strategy' : proof (symTrans q =><> r).
Proof.
  unfold proof.
  pcofix strategy'.
  pfold.
  apply Str with (dsj q r) ; auto.
  apply Spl.
  apply Stp ; auto.
  constructor.
Qed.

Lemma strategy : proof  (l =><> r).
Proof.
  pfold.
  apply Str with (dsj q r) ; auto.
  apply Spl.
  * apply Stp ; auto.
    left.    
    apply strategy'.
  *  constructor.
Qed.

End CompleteStrategy.


(* we now prove that a "q" like in CompleteStrategy always exists for valid rlfs *)
Definition coReach(r:SymState) : SymState :=
  fun s => ~ r s /\ forall tau, s = head tau-> isPath tau -> isFinite tau -> reach tau r.


Lemma init : forall f, valid f -> imp (lhs f) (dsj (coReach (rhs f)) (rhs f)).
Proof.
  unfold valid, coReach, imp, dsj; intros.
  destruct f as (l,r) ; cbn in *.
  destruct (classic (r s)) ; auto.
  left ; split; auto ; intros ; subst.
   apply H; auto.    
Qed.




Lemma nonBlock : forall r,  imp (cnj (coReach r) final) bot.
Proof.
  unfold imp, cnj, coReach, bot ; intros.
  do 2 destruct H.
  assert (isPath (one s)).
  * constructor ; auto.
  *  specialize (H1 (one s) (refl_equal _) H2).
     assert (Hfin : isFinite (one s)). {constructor. } 
     specialize (H1 Hfin).          
     pinversion H1. inversion H1; subst ; intuition.
     apply reach_gen_mon.
Qed.





Lemma quasiInv : forall r, imp (symTrans (coReach r)) (dsj  (coReach r) r).
Proof.
  unfold imp, symTrans, coReach, dsj; intros.
  destruct H as [s' [[Hnr Hreach] Htrans]].
  destruct (classic (r s)) ; auto.
  left  ; split ; auto ; intros ; subst.
  assert (reach (add s' tau) r).
  * apply Hreach ; try reflexivity ; constructor ; auto.
  * apply reach'_reach.
    apply reach_reach' in H0.
    pinversion H0 ; inversion H0 ; subst ; intuition.
      apply reach_gen'_mon.
Qed.

Theorem completeness : forall f,  valid f -> proof  f.
  intros.
  destruct f as (l,r).
  apply strategy with (coReach r).
  * apply (init (l =><> r))  ; auto.
  *  apply nonBlock.
  *  apply quasiInv.
Qed.

End LogicMod.